@@ -1,11 +1,11 @@
-//! Goose Eggs contains helpful functions and structures for writing
-//! [`Goose`](https://docs.rs/goose) load tests.
+//! Goose Eggs are helpful in writing [`Goose`](https://docs.rs/goose) load tests.
 use goose::goose::GooseResponse;
 use goose::prelude::*;
 use log::info;
+use regex::Regex;
 use reqwest::header::HeaderMap;
 
-use regex::Regex;
+pub mod drupal;
 
 /// Define one or more items to be validated in a web page response.
 ///
@@ -554,18 +554,105 @@ impl<'a> Header<'a> {
     }
 }
 
-/// Returns a [`bool`] indicating whether or not the title (case insensitive) is
-/// found within the html.
+/// Use a regular expression to get the HTML header from the web page.
 ///
-/// A valid title starts with `<title>foo` where `foo` is the expected title text.
-/// Returns [`true`] if the expected title is found, otherwise returns [`false`].
+/// # Example
+/// ```rust
+/// use goose_eggs::get_html_header;
 ///
-/// This function is case insensitive, if a title of "foo" is specified it will
-/// match "foo" or "Foo" or "FOO".
+/// // For this example we grab just a subset of a web page, enough to demonstrate. Normally
+/// // you'd use the entire html snippet returned from [`validate_and_load_static_assets`].
+/// let html = r#"
+/// <html lang="en" dir="ltr">
+///   <head>
+///     <meta charset="utf-8" />
+///     <link rel="canonical" href="https://example.com/" />
+///     <link rel="shortlink" href="https://example.com/" />
+///     <meta name="Generator" content="Drupal 9 (https://www.drupal.org)" />
+///     <meta name="MobileOptimized" content="width" />
+///     <meta name="HandheldFriendly" content="true" />
+///     <meta name="viewport" content="width=device-width, initial-scale=1.0" />
+///     <title>Example Website</title>
+///   </head>
+/// <body>
+///   This is the web page body.
+/// </body>
+/// </html>
+/// "#;
+///
+/// let html_header = get_html_header(html);
+/// assert!(!html_header.is_none());
+/// ```
+pub fn get_html_header(html: &str) -> Option<String> {
+    let re = Regex::new(r#"<head>(.*?)</head>"#).unwrap();
+    // Strip carriage returns to simplify regex.
+    let line = html.replace("\n", "");
+    // Return the entire html header, a subset of the received html.
+    re.captures(&line).map(|value| value[0].to_string())
+}
+
+/// Use a regular expression to get the web page title.
+///
+/// # Example
+/// ```rust
+/// use goose_eggs::{get_html_header, get_title};
+///
+/// // For this example we grab just a subset of a web page, enough to demonstrate. Normally
+/// // you'd use the entire html snippet returned from [`validate_and_load_static_assets`].
+/// let html = r#"
+/// <html lang="en" dir="ltr">
+///   <head>
+///     <meta charset="utf-8" />
+///     <link rel="canonical" href="https://example.com/" />
+///     <link rel="shortlink" href="https://example.com/" />
+///     <meta name="Generator" content="Drupal 9 (https://www.drupal.org)" />
+///     <meta name="MobileOptimized" content="width" />
+///     <meta name="HandheldFriendly" content="true" />
+///     <meta name="viewport" content="width=device-width, initial-scale=1.0" />
+///     <title>Example Website</title>
+///   </head>
+/// <body>
+///   This is the web page body.
+/// </body>
+/// </html>
+/// "#;
+///
+/// // Start by extracting the HTML header from the HTML.
+/// let html_header = get_html_header(html).map_or_else(|| "".to_string(), |h| h.to_string());
+/// // Next extract the title from the HTML header.
+/// let title = get_title(&html_header).map_or_else(|| "".to_string(), |t| t.to_string());
+/// assert_eq!(title, "Example Website");
+/// ```
+pub fn get_title(html: &str) -> Option<String> {
+    let re = Regex::new(r#"<title>(.*?)</title>"#).unwrap();
+    // Strip carriage returns to simplify regex.
+    let line = html.replace("\n", "");
+    // Return the entire title, a subset of the received html.
+    re.captures(&line).map(|value| value[1].to_string())
+}
+
+/// Returns a [`bool`] indicating whether or not the title (case insensitive) on the
+/// webpage contains the provided string.
 ///
 /// While you can invoke this function directly, it's generally preferred to invoke
 /// [`validate_and_load_static_assets`] which in turn invokes this function.
 ///
+/// A valid title is found between `<title></title>` tags inside `<head></head>` tags.
+/// For example, if the title is as follows:
+/// ```html
+/// <head>
+///   <title>this is the title</title>
+/// </head>
+/// ```
+///
+/// Then a call to `valid_title("the title")` will return [`true`], whereas a call to
+/// `valid_title("foo")` will return [`false`].
+///
+/// This function is case insensitive, so in the above example calling
+/// `valid_title("The Title")` and `valid_title("THE TITLE")` will both also return
+/// [`true`]. The function only tests if the title includes the specified text, the
+/// title can also include other text and will still be considered valid.
+///
 /// # Example
 /// ```rust
 /// use goose::prelude::*;
@@ -582,10 +669,11 @@ impl<'a> Header<'a> {
 ///             let headers = &response.headers().clone();
 ///             match response.text().await {
 ///                 Ok(html) => {
+///                     // Confirm that the HTML header includes the expected title.
 ///                     let title = "example";
 ///                     if !valid_title(&html, title) {
 ///                         return user.set_failure(
-///                             &format!("{}: title not found: {}", goose.request.url, title),
+///                             &format!("{}: title not found: {}", goose.request.raw.url, title),
 ///                             &mut goose.request,
 ///                             Some(&headers),
 ///                             Some(&html),
@@ -594,7 +682,7 @@ impl<'a> Header<'a> {
 ///                 }
 ///                 Err(e) => {
 ///                     return user.set_failure(
-///                         &format!("{}: failed to parse page: {}", goose.request.url, e),
+///                         &format!("{}: failed to parse page: {}", goose.request.raw.url, e),
 ///                         &mut goose.request,
 ///                         Some(&headers),
 ///                         None,
@@ -604,7 +692,7 @@ impl<'a> Header<'a> {
 ///         }
 ///         Err(e) => {
 ///             return user.set_failure(
-///                 &format!("{}: no response from server: {}", goose.request.url, e),
+///                 &format!("{}: no response from server: {}", goose.request.raw.url, e),
 ///                 &mut goose.request,
 ///                 None,
 ///                 None,
@@ -616,8 +704,14 @@ impl<'a> Header<'a> {
 /// }
 /// ```
 pub fn valid_title(html: &str, title: &str) -> bool {
-    html.to_ascii_lowercase()
-        .contains(&("<title>".to_string() + title.to_ascii_lowercase().as_str()))
+    // Extract the HTML header from the provided html.
+    let html_header = get_html_header(html).map_or_else(|| "".to_string(), |h| h);
+    // Next extract the title from the HTML header.
+    let html_title = get_title(&html_header).map_or_else(|| "".to_string(), |t| t);
+    // Finally, confirm that the title contains the expected text.
+    html_title
+        .to_ascii_lowercase()
+        .contains(title.to_ascii_lowercase().as_str())
 }
 
 /// Returns a [`bool`] indicating whether or not an arbitrary str (case sensitive) is found
@@ -650,7 +744,7 @@ pub fn valid_title(html: &str, title: &str) -> bool {
 ///                     let text = r#"<code class="language-console">$ cargo new hello_world --bin"#;
 ///                     if !valid_text(&html, text) {
 ///                         return user.set_failure(
-///                             &format!("{}: text not found: {}", goose.request.url, text),
+///                             &format!("{}: text not found: {}", goose.request.raw.url, text),
 ///                             &mut goose.request,
 ///                             Some(&headers),
 ///                             Some(&html),
@@ -659,7 +753,7 @@ pub fn valid_title(html: &str, title: &str) -> bool {
 ///                 }
 ///                 Err(e) => {
 ///                     return user.set_failure(
-///                         &format!("{}: failed to parse page: {}", goose.request.url, e),
+///                         &format!("{}: failed to parse page: {}", goose.request.raw.url, e),
 ///                         &mut goose.request,
 ///                         Some(&headers),
 ///                         None,
@@ -669,7 +763,7 @@ pub fn valid_title(html: &str, title: &str) -> bool {
 ///         }
 ///         Err(e) => {
 ///             return user.set_failure(
-///                 &format!("{}: no response from server: {}", goose.request.url, e),
+///                 &format!("{}: no response from server: {}", goose.request.raw.url, e),
 ///                 &mut goose.request,
 ///                 None,
 ///                 None,
@@ -707,7 +801,7 @@ pub fn valid_text(html: &str, text: &str) -> bool {
 ///             let headers = &response.headers().clone();
 ///             if !header_is_set(headers, &Header::name("server")) {
 ///                 return user.set_failure(
-///                     &format!("{}: header not found: {}", goose.request.url, "server"),
+///                     &format!("{}: header not found: {}", goose.request.raw.url, "server"),
 ///                     &mut goose.request,
 ///                     Some(&headers),
 ///                     None,
@@ -716,7 +810,7 @@ pub fn valid_text(html: &str, text: &str) -> bool {
 ///         }
 ///         Err(e) => {
 ///             return user.set_failure(
-///                 &format!("{}: no response from server: {}", goose.request.url, e),
+///                 &format!("{}: no response from server: {}", goose.request.raw.url, e),
 ///                 &mut goose.request,
 ///                 None,
 ///                 None,
@@ -754,7 +848,7 @@ pub fn header_is_set(headers: &HeaderMap, header: &Header) -> bool {
 ///             let headers = &response.headers().clone();
 ///             if !valid_header_value(headers, &Header::name_value("server", "nginx")) {
 ///                 return user.set_failure(
-///                     &format!("{}: server header value not correct: {}", goose.request.url, "nginx"),
+///                     &format!("{}: server header value not correct: {}", goose.request.raw.url, "nginx"),
 ///                     &mut goose.request,
 ///                     Some(&headers),
 ///                     None,
@@ -763,7 +857,7 @@ pub fn header_is_set(headers: &HeaderMap, header: &Header) -> bool {
 ///         }
 ///         Err(e) => {
 ///             return user.set_failure(
-///                 &format!("{}: no response from server: {}", goose.request.url, e),
+///                 &format!("{}: no response from server: {}", goose.request.raw.url, e),
 ///                 &mut goose.request,
 ///                 None,
 ///                 None,
@@ -824,11 +918,11 @@ pub fn valid_header_value(headers: &HeaderMap, header: &Header) -> bool {
 ///             match response.text().await {
 ///                 Ok(html) => {
 ///                     // Load all static elements on page.
-///                     load_static_elements(user, &html);
+///                     load_static_elements(user, &html).await;
 ///                 }
 ///                 Err(e) => {
 ///                     return user.set_failure(
-///                         &format!("{}: failed to parse page: {}", goose.request.url, e),
+///                         &format!("{}: failed to parse page: {}", goose.request.raw.url, e),
 ///                         &mut goose.request,
 ///                         Some(&headers),
 ///                         None,
@@ -838,7 +932,7 @@ pub fn valid_header_value(headers: &HeaderMap, header: &Header) -> bool {
 ///         }
 ///         Err(e) => {
 ///             return user.set_failure(
-///                 &format!("{}: no response from server: {}", goose.request.url, e),
+///                 &format!("{}: no response from server: {}", goose.request.raw.url, e),
 ///                 &mut goose.request,
 ///                 None,
 ///                 None,
@@ -874,10 +968,15 @@ pub async fn load_static_elements(user: &GooseUser, html: &str) {
     }
 }
 
-/// Validate the HTML response then extract and load all static elements on the page.
+/// Validate the HTML response, extract and load all static elements on the page, and
+/// return the HTML body.
 ///
 /// What is validated is defined with the [`Validate`] structure.
 ///
+/// If the page doesn't load, an empty String will be returned. If the page does load
+/// but validation fails, an Error is returned. If the page loads and there are no
+/// errors the body is returned as a String.
+///
 /// # Example
 /// ```rust
 /// use goose::prelude::*;
@@ -901,98 +1000,127 @@ pub async fn validate_and_load_static_assets<'a>(
     user: &GooseUser,
     mut goose: GooseResponse,
     validate: &'a Validate<'a>,
-) -> GooseTaskResult {
+) -> Result<String, GooseTaskError> {
+    let empty = "".to_string();
     match goose.response {
         Ok(response) => {
-            // Copy the headers so we have them for logging if there are errors.
-            let headers = &response.headers().clone();
+            // Validate status code if defined.
             let response_status = response.status();
+            if let Some(status) = validate.status {
+                if response_status != status {
+                    // Get as much as we can from the response for useful debug logging.
+                    let headers = &response.headers().clone();
+                    let html = response.text().await.unwrap_or_else(|_| "".to_string());
+                    user.set_failure(
+                        &format!(
+                            "{}: response status != {}]: {}",
+                            goose.request.raw.url, status, response_status
+                        ),
+                        &mut goose.request,
+                        Some(&headers),
+                        Some(&html),
+                    )?;
+                    // Exit as soon as validation fails, to avoid cascades of
+                    // errors whe na page fails to load.
+                    return Ok(html);
+                }
+            }
+
+            // Validate headers if defined.
+            let headers = &response.headers().clone();
+            for header in &validate.headers {
+                if !header_is_set(headers, header) {
+                    // Get as much as we can from the response for useful debug logging.
+                    let html = response.text().await.unwrap_or_else(|_| "".to_string());
+                    user.set_failure(
+                        &format!(
+                            "{}: header not included in response: {:?}",
+                            goose.request.raw.url, header
+                        ),
+                        &mut goose.request,
+                        Some(&headers),
+                        Some(&html),
+                    )?;
+                    // Exit as soon as validation fails, to avoid cascades of
+                    // errors whe na page fails to load.
+                    return Ok(html);
+                }
+                if let Some(h) = header.value {
+                    if !valid_header_value(headers, header) {
+                        // Get as much as we can from the response for useful debug logging.
+                        let html = response.text().await.unwrap_or_else(|_| "".to_string());
+                        user.set_failure(
+                            &format!(
+                                "{}: header does not contain expected value: {:?}",
+                                goose.request.raw.url, h
+                            ),
+                            &mut goose.request,
+                            Some(&headers),
+                            Some(&html),
+                        )?;
+                        // Exit as soon as validation fails, to avoid cascades of
+                        // errors whe na page fails to load.
+                        return Ok(html);
+                    }
+                }
+            }
+
+            // Extract the response body to validate and load static elements.
             match response.text().await {
                 Ok(html) => {
-                    // Validate status code if defined.
-                    if let Some(status) = validate.status {
-                        if response_status != status {
-                            return user.set_failure(
-                                &format!(
-                                    "{}: response status != {}]: {}",
-                                    goose.request.url, status, response_status
-                                ),
-                                &mut goose.request,
-                                Some(&headers),
-                                Some(&html),
-                            );
-                        }
-                    }
                     // Validate title if defined.
                     if let Some(title) = validate.title {
                         if !valid_title(&html, &title) {
-                            return user.set_failure(
-                                &format!("{}: title not found: {}", goose.request.url, title),
+                            user.set_failure(
+                                &format!("{}: title not found: {}", goose.request.raw.url, title),
                                 &mut goose.request,
                                 Some(&headers),
                                 Some(&html),
-                            );
+                            )?;
+                            // Exit as soon as validation fails, to avoid cascades of
+                            // errors whe na page fails to load.
+                            return Ok(html);
                         }
                     }
                     // Validate texts in body if defined.
                     for text in &validate.texts {
                         if !valid_text(&html, text) {
-                            return user.set_failure(
-                                &format!("{}: text not found on page: {}", goose.request.url, text),
-                                &mut goose.request,
-                                Some(&headers),
-                                Some(&html),
-                            );
-                        }
-                    }
-                    // Validate headers if defined.
-                    for header in &validate.headers {
-                        if !header_is_set(headers, header) {
-                            return user.set_failure(
+                            user.set_failure(
                                 &format!(
-                                    "{}: header not included in response: {:?}",
-                                    goose.request.url, header
+                                    "{}: text not found on page: {}",
+                                    goose.request.raw.url, text
                                 ),
                                 &mut goose.request,
                                 Some(&headers),
                                 Some(&html),
-                            );
-                        }
-                        if let Some(h) = header.value {
-                            if !valid_header_value(headers, header) {
-                                return user.set_failure(
-                                    &format!(
-                                        "{}: header does not contain expected value: {:?}",
-                                        goose.request.url, h
-                                    ),
-                                    &mut goose.request,
-                                    Some(&headers),
-                                    Some(&html),
-                                );
-                            }
+                            )?;
+                            // Exit as soon as validation fails, to avoid cascades of
+                            // errors whe na page fails to load.
+                            return Ok(html);
                         }
                     }
                     load_static_elements(user, &html).await;
+                    Ok(html)
                 }
                 Err(e) => {
-                    return user.set_failure(
-                        &format!("{}: failed to parse page: {}", goose.request.url, e),
+                    user.set_failure(
+                        &format!("{}: failed to parse page: {}", goose.request.raw.url, e),
                         &mut goose.request,
                         Some(&headers),
                         None,
-                    );
+                    )?;
+                    Ok(empty)
                 }
             }
         }
         Err(e) => {
-            return user.set_failure(
-                &format!("{}: no response from server: {}", goose.request.url, e),
+            user.set_failure(
+                &format!("{}: no response from server: {}", goose.request.raw.url, e),
                 &mut goose.request,
                 None,
                 None,
-            );
+            )?;
+            Ok(empty)
         }
     }
-
-    Ok(())
 }
